C This program computes various solutions to the 4th-corner problem.
C
C Reference: Legendre, P., R. Galzin & M. L. Harmelin-Vivien. 1997.
C     Relating behavior to habitat: solutions to the fourth-corner
C     problem. Ecology 78(2): 547-562.
C
C Macintosh adaptation and development upon a program written by PL for CDC 
C in June 1989. The data can be represented as follows: A(n1,n2) B(n1,n4)
C                                                       C(n3,n2) D(n3xn4)
C The problem consists in finding D, given A, B and C.
C
C Computing efficiency has sometimes been sacrificed to improve the clarity
C of the code. Users are welcome to experiment with their 4th-corner ideas
C by modifying the code, as long as publications give proper credit to this
C program and to the Ecology paper cited above.
C
C                             Pierre Legendre, September 1994 / May 1996
C ----------------------------------------------------------------------
C This version compiles and runs on PC-compatible computers using the 
C Microsoft Fortran as well as other compilers.
C ----------------------------------------------------------------------
C The line of Parameters below has to be repeated in almost every subroutine 
C of the program. Alternatively, it could be put in a separate file and called  
C by an 'Include' statement everywhere it is presently found. Of course, the 
C compiler would then have to be told that the material to be compiled consists
C of two files, i.e. the present file and the one with the Parameter statement.
C
C2345678901234567890123456789012345678901234567890123456789012345678901234567890
      Parameter (nn1=300,nn2=30,nn34=30)
      Integer EspD(nn34,nn34),n1,n2,n3,n4,DLT(nn34,nn34),DEQ(nn34,nn34),
     +        permod
      Real*8 mat(nn34)
      Real*8 A(nn1,nn2),B(nn1,nn34),C(nn34,nn2),D(nn34,nn34),tot,
     +       Z(nn34,nn34),DD(nn34,nn34),Dmin(nn34,nn34),Dmoy(nn34,nn34),
     +       Dmax(nn34,nn34),Prob(nn34),Dligne(nn34),Dcol(nn34),
     +       count(nn34,nn34),dseed
      Common A,B,C,D,tot,Z,DD,Dmin,Dmoy,Dmax,Prob,Dligne,Dcol,count,
     +       dseed,EspD,n1,n2,n3,n4,DLT,DEQ,permod,
     +       methode,nperm,nclasse,langue
C      Character tt*10
      Character*79 namea
C Dialog language: 1 = francais, other integer value = English
      langue=2
C Read data; computation method; permutation model; number of permutations
	  dseed=123457.0d0
      call ReadData
    2 if(langue.eq.1) then
         write(*,102) 
         else
         write(*,202) 
         endif
      read(*,*) permod
      if((permod.lt.0).or.(permod.gt.4)) then
         if(langue.eq.1) then
            write(*,*) 'Mauvaise reponse. Reprenons...'
            else
            write(*,*) 'Wrong answer. Try again...'
            endif
         goto 2
         endif
      if(permod.eq.0) then
         if(langue.eq.1) then
            stop 'Fin du programme.'
            else
            stop 'End of the program.'
            endif
         endif
      write(*,*)
      if(langue.eq.1) then
         write(*,*) 'Combien de permutations? (ex. 99, 999, 9999...)'
         else
         write(*,*) 'How many permutations? (ex. 99, 999, 9999...)'
         endif
      read(*,*) nperm
      if(nperm.lt.0) nperm=0
      if(langue.eq.1) then
         write(8,101) methode,permod,nperm
         else
         write(8,201) methode,permod,nperm
         endif
      write(*,*)
C Time elapsed, using g77's intrinsic ftime subroutine
C (it doesn't seem to work, though...)
      write(*,109) ftime
C Time elapsed, using Microsoft subroutine GETTIM
C     call gettim(ihr,imin,isec,i100th)
C     write(*,109) ihr1,imin1,isec1
C This is the corresponding LSFortran code (Macintosh)
C      time(tt)
C      write(*,*) tt
C      debut=secnds(0.0)
C
C Going to the main calculation subroutine
C
      call Compute
C
      write(*,109) ftime
C      call gettim(ihr1,imin1,isec1,i100th1)
C     write(*,109) ihr1,imin1,isec1
C      call time(tt)
C      write(6,*) tt
C      fin=secnds(debut)
C Calculating time elapsed
      fin=(float(i100th1-i100th)/100.0)+float(isec1-isec)+
     +    60.0*float(imin1-imin)+3600.0*float(ihr1-ihr)
      if(langue.eq.1) then
         write(6,108) fin
         write(8,108) fin
         else
         write(6,208) fin
         write(8,208) fin
         endif
      write(*,*)
C Write inflated matrix to a file?
   30 if(langue.eq.1) then
         write(*,*) 'Inscrire la matrice soufflee dans un fichier?'
         write(*,*) '(0) Non, (1) Oui.'
         else
         write(*,*) 'Write the inflated data matrix to a file?'
         write(*,*) '(0) No, (1) Yes.'
         endif
      read(*,*) inscrire
      if((inscrire.lt.0).or.(inscrire.gt.1)) goto 30
      if(inscrire.eq.0) goto 40
      if(langue.eq.1) then
         write(*,*) 'Nom du fichier contenant la matrice soufflee?'
         read(*,'(A)') namea
         open(7,File=namea,Status='NEW')
         else
         write(*,*) 'Name of the file containing the inflated matrix?'
         read(*,'(A)') namea
         open(7,File=namea,Status='NEW')
         endif
C      write(*,*) namea
      ligne=0
      do 36 i=1,n1
      do 36 j=1,n2
      if(A(i,j).eq.1.0) then
         ligne=ligne+1
         col=0
         do 32 k1=1,n3
         col=col+1
   32    mat(col)=C(k1,j)
         do 34 k2=1,n4
         col=col+1
   34    mat(col)=B(i,k2)
         write(7,103) (mat(k), k=1,col)
         endif
   36 continue
      if(langue.eq.1) then
         write(*,100) ligne,n3,n4
         else
         write(*,200) ligne,n3,n4
         endif
C End of calculations
   40    write(6,*)
      if(langue.eq.1) then
         write(6,*) 'Fin du programme.'
         else
         write(6,*) 'End of the program.'
         endif
      stop
  100 format(i6,' lignes; C = ',i2,
     +       ' premieres col., B = ',i2,' dernieres col.')
  200 format(i6,' rows; C = first ',i2,' col., B = last ',i2,' col.')
  101 format(' Calcul: methode',i2,'.  Permutations: methode',i2,
     +       '.  N. permutations =',i5,'.')
  201 format(' Computing method',i2,'.  Permutation method',i2,
     +       '.  N. permutations =',i5,'.')
  102 format(
     +' Modele de permutation? (1 a 4; voir l''article Ecology)'/
     +' (1) Permuter les colonnes separement pour chaque ligne de A'/
     +' (2) Permuter les colonnes entieres de A'/
     +' (3) Permuter les lignes separement pour chaque colonne de A'/
     +' (4) Permuter les lignes entieres de A'/
     +' (0) Quitter le programme.')
  202 format(
     +' Permutation model? (1 to 4; see Ecology paper)'/
     +' (1) Permute columns, independently within each row of A'/
     +' (2) Permute whole columns of A at random'/
     +' (3) Permute rows, independently within each column of A'/
     +' (4) Permute whole rows of A at random'/
     +' (0) Quit the program.')
  103 format(10f8.2)
  108 format(' Duree reelle du calcul:',f10.2,' secondes')
  208 format(' Real time spent:',f10.2,' seconds')
  109 format(' Time:',i2.2,':',i2.2,':',i2.2)
      end
      
      Subroutine ReadData
      Parameter (nn1=300,nn2=30,nn34=30)
      Integer EspD(nn34,nn34),n1,n2,n3,n4,DLT(nn34,nn34),DEQ(nn34,nn34),
     +        permod
      Real*8 A(nn1,nn2),B(nn1,nn34),C(nn34,nn2),D(nn34,nn34),tot,
     +       Z(nn34,nn34),DD(nn34,nn34),Dmin(nn34,nn34),Dmoy(nn34,nn34),
     +       Dmax(nn34,nn34),Prob(nn34),Dligne(nn34),Dcol(nn34),
     +       count(nn34,nn34),dseed
      Common A,B,C,D,tot,Z,DD,Dmin,Dmoy,Dmax,Prob,Dligne,Dcol,count,
     +       dseed,EspD,n1,n2,n3,n4,DLT,DEQ,permod,
     +       methode,nperm,nclasse,langue
C      Character nom*10
      Character*79 namea
      nclasse=0
      if(langue.eq.1) then
         write(*,100)
         else
         write(*,200)
         endif
C What computation method?
    2 if(langue.eq.1) then
         write(*,101)
         else
         write(*,201)
         endif
      read(*,*) methode
      if((methode.lt.0).or.(methode.gt.6)) then
         if(langue.eq.1) then
            write(*,*) 'Mauvaise reponse. Reprenons...'
            else
            write(*,*) 'Wrong answer. Try again...'
            endif
         goto 2
         endif
      if(methode.eq.0) then
         if(langue.eq.1) then
            stop 'Fin du programme.'
            else
            stop 'End of the program.'
            endif
         endif
C Size of the matrices?  
    4 if(langue.eq.1) then
         write(*,*)
         write(*,*) 'Tapez n1, n2, n3, n4.'
         else
         write(*,*) 'Type n1, n2, n3, n4.'
         endif
      read(*,*) n1,n2,n3,n4
      write(*,*)
      if((n1.gt.nn1).or.(n2.gt.nn2).or.(n3.gt.nn34).or.(n4.gt.nn34))then
         if(langue.eq.1) then
            stop 'Matrice trop grande. Redimensionnez le programme.'
            else
            stop 'Matrix too large. Redimension the program.'
            endif
         endif
      if((n1.lt.0).or.(n2.lt.0).or.(n3.lt.0).or.(n4.lt.0)) then 
         if(langue.eq.1) then
            write(*,*) 'Mauvaise reponse: dimension <= 0. Reprenons...'
            else
            write(*,*) 'Wrong answer: dimension <= 0. Try again...'
            endif
         goto 4
         endif
C Output file?
      if(langue.eq.1) then
         write(*,*) 'Nom du fichier des resultats ?'
         read(*,'(A)') namea
         open(8,File=namea,Status='UNKNOWN')
         else
         write(*,*) 'Name to be given to the file of results ?'
         read(*,'(A)') namea
         open(8,File=namea,Status='UNKNOWN')
         endif
C      write(*,*) namea
C      write(*,*)
C Read in the data tables --
C Matrix A:
      if(langue.eq.1) then
         write(*,*) 'Nom du fichier de donnees A (n1xn2) ?'
         else
         write(*,*) 'Name of input data file A (n1xn2) ?'
         endif
      read(*,'(A)') namea
      open(1,File=namea,Status='OLD')
C      write(*,*) namea
      if(langue.eq.1) then
         write(8,103) 'A',namea
         else
         write(8,203) 'A',namea
         endif
      do 12 i=1,n1
   12 read(1,*) (A(i,j), j=1,n2)
C      nom=' A =      '
C      call Imprime(A,nn1,nn2,n1,n2,nom)
C Matrix B:
      if(langue.eq.1) then
         write(*,*) 'Nom du fichier de donnees B (n1xn4) ?'
         else
         write(*,*) 'Name of input data file B (n1xn4) ?'
         endif
      read(*,'(A)') namea
      open(2,File=namea,Status='OLD')
C      write(*,*) namea
      if(langue.eq.1) then
         write(8,103) 'B',namea
         else
         write(8,203) 'B',namea
         endif
      do 14 i=1,n1
   14 read(2,*) (B(i,j), j=1,n4)
      if(methode.eq.5) then
         do 16 i=1,n1
         mm=B(i,1)
         if(mm.gt.nclasse) nclasse=mm
   16    continue
         write(*,102) 'B',nclasse
         write(8,102) 'B',nclasse
         if(nclasse.gt.nn34) then
            if(langue.eq.1) then
            stop 'Trop de classes dans B. Redimensionnez le programme.'
               else
               stop 'Too many classes in B. Redimension the program.'
               endif
			endif
         endif
C      nom=' B =      '
C      call Imprime(B,nn1,nn34,n1,n4,nom)
C Matrix C:
      if(langue.eq.1) then
         write(*,*) 'Nom du fichier de donnees C (n3xn2) ?'
         else
         write(*,*) 'Name of input data file C (n3xn2) ?'
         endif
      read(*,'(A)') namea
      open(3,File=namea,Status='OLD')
C      write(*,*) namea
      if(langue.eq.1) then
         write(8,103) 'C',namea
         else
         write(8,203) 'C',namea
         endif
      write(8,*)
      do 18 i=1,n3
   18 read(3,*) (C(i,j), j=1,n2)
      if(methode.eq.6) then
         do 20 j=1,n2
         mm=C(1,j)
         if(mm.gt.nclasse) nclasse=mm
   20    continue
         write(*,102) 'C',nclasse
         write(8,102) 'C',nclasse
         if(nclasse.gt.nn34) then
            if(langue.eq.1) then
            stop 'Trop de classes dans C. Redimensionnez le programme.'
               else
               stop 'Too many classes in C. Redimension the program.'
               endif
			endif
         endif
C      nom=' C =      '
C      call Imprime(C,nn34,nn2,n3,n2,nom)
      if(langue.eq.1) then
         write(*,*)'Fin de la lecture des matrices.'
         else
         write(*,*)'The input data matrices have been read.'
         endif
      write(*,*)
      return
  100 format(' Calcul des statistiques du quatrieme coin',
     +' (Legendre et al., Ecology, 1996.).'//
     +' Les donnees doivent etre presentees comme suit:'//
     +' Matrice A (n1xn2): lignes = esp., col. = stat.; 0/1 seulement.'/
     +' Matrice B (n1xn4): lignes = especes, col. = comportements.'/
     +' Matrice C (n3xn2): lignes = var. d''habitat, col. = stations.'/)
  200 format(
     +' Computing fourth-corner statistics',
     +' (Legendre et al., Ecology, 1996).'//
     +' The data files must be presented as follows:'//
     +' Matrix A (n1xn2): rows = species, col. = stations; 0/1 only.'/
     +' Matrix B (n1xn4): rows = species, col. = behaviours.'/
     +' Matrix C (n3xn2): rows = habitat var., col. = stations.'/)
  101 format(
     +' Choisir une methode de calcul:'/
     +' (1) B et C = var. nominales (codage disjonctif): contingence'/
     +' (2) Correlations entre B et C (plusieurs var. quantitatives)'/
     +' (3) F: B= nominale (codage disjonctif), C= 1 var. quantitative'/
     +' (4) F: B= 1 var. quantitative, C= nominale (codage disjonctif)'/
     +' (5) F: B= 1 var. nominale (1 colonne), C= 1 var. quantitative'/
     +' (6) F: B= 1 var. quantitative, C= 1 var. nominale (1 ligne)'/
     +' (0) Quitter le programme.')
  201 format(
     +' Choose a computation method:'/
     +' (1) B and C = nominal var. (disjunctive coding): contingency'/
     +' (2) Correlations between B and C (several quantitative var.)'/
     +' (3) F: B= nominal (disjunctive coding), C= 1 quantitative var.'/
     +' (4) F: B= 1 quantitative var., C= nominal (disjunctive coding)'/
     +' (5) F: B= 1 nominal var. (1 column), C= 1 quantitative var.'/
     +' (6) F: B= 1 quantitative var., C= 1 nominal var. (1 row only)'/
     +' (0) Quit the program.')
  102 format(' Variable ',a1,':',i3,' classes')
  103 format(' Matrice ',A1,' = ',A32)
  203 format(' Matrix ',A1,' = ',A32)
      end

      Subroutine Compute
      Parameter (nn1=300,nn2=30,nn34=30)
      Integer EspD(nn34,nn34),n1,n2,n3,n4,DLT(nn34,nn34),DEQ(nn34,nn34),
     +        permod
      Integer iordre(nn1)
      Real*8 Aprime(nn2,nn1)
      Real*8 G,Chi2P,GG,Chi2PP,F,FF,probg,probchi,probf
      Real*8 A(nn1,nn2),B(nn1,nn34),C(nn34,nn2),D(nn34,nn34),tot,
     +       Z(nn34,nn34),DD(nn34,nn34),Dmin(nn34,nn34),Dmoy(nn34,nn34),
     +       Dmax(nn34,nn34),Prob(nn34),Dligne(nn34),Dcol(nn34),
     +       count(nn34,nn34),dseed
      Common A,B,C,D,tot,Z,DD,Dmin,Dmoy,Dmax,Prob,Dligne,Dcol,count,
     +       dseed,EspD,n1,n2,n3,n4,DLT,DEQ,permod,
     +       methode,nperm,nclasse,langue
      Common /pourperm/ Aprime,iordre
C Vector IORDRE must be dimensioned to the largest of (nn1,nn2)
      nval=n2
      if(permod.ge.3) nval=n1
      do 2 i=1,nval
    2 iordre(i)=i
C Calculate: D = C A' B
      if(methode.eq.1) then
         LTG=0
         IEQG=1
         LTCHI2P=0
         IEQCHI2=1
         endif
      if(methode.ge.3) then
         LTF=0
         IEQF=1
         if(langue.eq.1) then
            write(8,123)
            else
            write(8,223)
            endif
         else
         if(langue.eq.1) then
            write(8,122)
            else
            write(8,222)
            endif
         endif
      if(langue.eq.1) then
         write(8,124)
         else
         write(8,224)
         endif
      tot=0.0
      do 12 i=1,n1
      do 12 j=1,n2
   12 tot=tot+A(i,j)
      if(langue.eq.1) then
         write(8,100) tot
         else
         write(8,200) tot
         endif
      if(methode.eq.1) then
            call Prod1(0)
            call Test(G,Chi2P,0)
         else if(methode.eq.2) then
            call Prod1(0)
         else
            call Prod2(0,F)
         endif
      if(methode.lt.5) then
         nD3=n3
         nD4=n4
         else if(methode.eq.5) then
         nD3=1
         nD4=nclasse
         else
         nD3=nclasse
         nD4=1
         endif
      if(methode.ge.3) then
         do 14 i=1,nD3
         do 14 j=1,nD4
   14    EspD(i,j)=1
         endif
C Begin permutations
      do 16 i=1,nD3
      do 16 j=1,nD4
      Dmin(i,j)=D(i,j)
      Dmoy(i,j)=D(i,j)
      Dmax(i,j)=D(i,j)
      DLT(i,j)=0
   16 DEQ(i,j)=1
      write(*,101)
      do 30 k=1,nperm
      if(nperm.le.1000) then
         if(mod(k,100).eq.0) write(*,102) k
         else
         if(mod(k,1000).eq.0) write(*,102) k
         endif
      if(methode.eq.1) then
            call Prod1(1)
            call Test(GG,Chi2PP,1)
            if(GG.lt.G) LTG=LTG+1
            if(GG.eq.G) IEQG=IEQG+1
            if(Chi2PP.lt.Chi2P) LTCHI2P=LTCHI2P+1
            if(Chi2PP.eq.Chi2P) IEQCHI2=IEQCHI2+1
         else if(methode.eq.2) then
            call Prod1(1)
         else
            call Prod2(1,FF)
            if(FF.lt.F) LTF=LTF+1
            if(FF.eq.F) IEQF=IEQF+1
         endif
      do 22 i=1,nD3
      do 20 j=1,nD4
      if(methode.ge.3) then
         if(count(nD3,nD4).le.1) then
            goto 20
            else
            EspD(i,j)=EspD(i,j)+1
            endif
         endif
      if(Dmin(i,j).gt.DD(i,j)) Dmin(i,j)=DD(i,j)
      Dmoy(i,j)=Dmoy(i,j)+DD(i,j)
      if(Dmax(i,j).lt.DD(i,j)) Dmax(i,j)=DD(i,j)
      if(DD(i,j).lt.D(i,j))    DLT(i,j)=DLT(i,j)+1
      if(DD(i,j).eq.D(i,j))    DEQ(i,j)=DEQ(i,j)+1
   20 continue
   22 continue
   30 continue
      write(*,102) k-1
C End permutations
C
C Print table of statistics
      do 36 i=1,nD3
      if(nperm.eq.0) goto 34
      do 32 j=1,nD4
      if(methode.le.2) then
         Dmoy(i,j)=Dmoy(i,j)/dfloat(nperm+1)
         else
         Dmoy(i,j)=Dmoy(i,j)/dfloat(EspD(i,j))
         endif
C                            One-tailed test on the left or right?
      if(methode.eq.1) then
C                            Criterion with method 1: Dmoy
         if(D(i,j).lt.Dmoy(i,j)) then
            Prob(j)=dfloat(DLT(i,j)+DEQ(i,j))/dfloat(nperm+1)
            else
            Prob(j)=1.0-dfloat(DLT(i,j))/dfloat(nperm+1)
            endif      
        else if(methode.eq.2) then
C                            Criterion with method 2: Signe de r
         if(D(i,j).lt.0.0) then
            Prob(j)=dfloat(DLT(i,j)+DEQ(i,j))/dfloat(nperm+1)
            else
            Prob(j)=1.0-dfloat(DLT(i,j))/dfloat(nperm+1)
            endif
        else
C                            Methods 3 to 6: test always in left-hand tail
         Prob(j)=dfloat(DLT(i,j)+DEQ(i,j))/dfloat(EspD(i,j))
        endif
   32 continue
   34 write(8,*)
      if(methode.eq.2) then
         write(8,106) 'r(i,j)', (D(i,j),    j=1,nD4)
         else
         write(8,106) 'D(i,j)', (D(i,j),    j=1,nD4)
         endif
      if(nperm.eq.0) goto 36
C      if(methode.eq.1) then
C         write(8,107) 'Esp(D)', (EspD(i,j), j=1,nD4)
C         endif
      write(8,106)    'Min   ', (Dmin(i,j), j=1,nD4)
      write(8,106)    'Moy   ', (Dmoy(i,j), j=1,nD4)
      write(8,106)    'Max   ', (Dmax(i,j), j=1,nD4)
      write(8,107)    'N. LT ', (DLT(i,j),  j=1,nD4)
      write(8,107)    'N. EQ ', (DEQ(i,j),  j=1,nD4)
      if(methode.ge.3) then
         write(8,107)'N. GT ',((EspD(i,j)-DLT(i,j)-DEQ(i,j)),j=1,nD4)
         endif
      write(8,106)    'Prob  ', (Prob(j),   j=1,nD4)
   36 continue
C Print global statistics, G or F
      write(8,*)
      if(methode.eq.1) then
         idf=(n3-1)*(n4-1)
         probg  = 1.0 - dfloat(LTG)    /dfloat(nperm+1)
         probchi= 1.0 - dfloat(LTCHI2P)/dfloat(nperm+1)
         if(nperm.eq.0) then
            write(8,112) G    ,idf
            write(8,113) CHI2P,idf
            else
            write(8,114) G    ,idf,nperm,probg
            write(8,115) CHI2P,idf,nperm,probchi
            endif
C Computing normal deviates for the cells of the contingency table
C         if(langue.eq.1) then
C            write(8,120)
C            else
C            write(8,220)
C            endif
C         do 48 i=1,n3
C   48    write(8,121) (Z(i,j), j=1,n4)
         write(8,*)
        else if(methode.ge.3) then
         probf = 1.0 - dfloat(LTF)/dfloat(nperm+1)
         write(8,116) F,nperm,probf,LTF,IEQF
        endif
C
      return
  100 format(' Nombre total d''observations =',f7.0)
  200 format(' Total number of observations =',f7.0)
  101 format(/' N. permutations =      0')
  102 format('+N. permutations =',i7)
  106 format(1x,a6,7f10.5,10(/7x,7f10.5))
  107 format(1x,a6,7i10,10(/7x,7i10))
  112 format(' G    = ',f10.5,' DF = ',i4)
  113 format(' Chi2 = ',f10.5,' DF = ',i4)
  114 format(' G    = ',f10.5,' DF = ',i4,' Prob.(',i5,' perm.) =',f9.5)
  115 format(' Chi2 = ',f10.5,' DF = ',i4,' Prob.(',i5,' perm.) =',f9.5)
  116 format(' F = ',f10.5,'   Prob. (',i5,' perm.) =',f9.5/
     +       ' lt(F) =',i5,'   eq(F) =',i5)
  120 format(/' Tableau des ecarts normaux "Z" (Neu et al. 1974)'/)
  220 format(/' Table of "Z" normal deviates (Neu et al. 1974)'/)
  121 format(1x,7f10.5)
  122 format(
     +' Tests unilateral dans la queue de la statistique D(i,j).')
  222 format(
     +' One-tailed test in tail of statistic D(i,j).')
  123 format(
     +' D(i,j) = statistique d''homogeneite = Var(intra)/Var(totale),'/
     +' test unilateral dans la queue de gauche. Si une permutation'/
     +' produit 0 ou 1 observation dans un groupe, l''homogeneite ne'/
     +' peut etre calculee; ce resultat est elimine des permutations.')
  223 format(
     +' D(i,j) = homogeneity statistic = Var(within)/Var(total),'/
     +' one-tailed test in left-hand tail. If a permutation produces'/
     +' 0 or 1 observation in a group the homogeneity statistic cannot'/
     +' be computed; this result is eliminated from the permutations.')
  124 format(
     +' Probabilites incluent la correction de Hope (la vraie dans EQ)')
  224 format(
     +' Probabilities include the Hope correction (true value in EQ).')
      end
      
      Subroutine Permute
      Parameter (nn1=300,nn2=30,nn34=30)
      Integer EspD(nn34,nn34),n1,n2,n3,n4,DLT(nn34,nn34),DEQ(nn34,nn34),
     +        permod
      Integer iordre(nn1)
      Real*8 Aprime(nn2,nn1)
      Real*8 A(nn1,nn2),B(nn1,nn34),C(nn34,nn2),D(nn34,nn34),tot,
     +       Z(nn34,nn34),DD(nn34,nn34),Dmin(nn34,nn34),Dmoy(nn34,nn34),
     +       Dmax(nn34,nn34),Prob(nn34),Dligne(nn34),Dcol(nn34),
     +       count(nn34,nn34),dseed
      Common A,B,C,D,tot,Z,DD,Dmin,Dmoy,Dmax,Prob,Dligne,Dcol,count,
     +       dseed,EspD,n1,n2,n3,n4,DLT,DEQ,permod,
     +       methode,nperm,nclasse,langue
      Common /pourperm/ Aprime,iordre
      Character nom*10
      nom=' Aprime = '
      goto (10,20,30,40) permod
C
C Permutations: Model 1 -- Permute columns, independently within each row of A
   10 do 12 i=1,n1
      call ggper2(dseed,n2)
      do 12 j=1,n2
   12 Aprime(j,i)=A(i,iordre(j))
      return
C
C Permutations: Model 2 -- Permute whole columns of A at random
   20 call ggper2(dseed,n2)
      do 22 i=1,n1
      do 22 j=1,n2
   22 Aprime(j,i)=A(i,iordre(j))
      return
C
C Permutations: Model 3 -- Permute rows, independently within each column of A
   30 do 32 j=1,n2
      call ggper2(dseed,n1)
      do 32 i=1,n1
   32 Aprime(j,i)=A(iordre(i),j)
      return
C
C Permutations: Model 4 -- Permute whole rows of A at random
   40 call ggper2(dseed,n1)
      do 42 j=1,n2
      do 42 i=1,n1
   42 Aprime(j,i)=A(iordre(i),j)
      return
      end

      Subroutine Prod1(ipar)
      Parameter (nn1=300,nn2=30,nn34=30)
      Integer EspD(nn34,nn34),n1,n2,n3,n4,DLT(nn34,nn34),DEQ(nn34,nn34),
     +        permod
      Integer iordre(nn1)
      Real*8 Aprime(nn2,nn1),Caprime(nn34,nn1),
     +       bbar(nn34),sb(nn34),sb2(nn34),cbar(nn34),sc(nn34),sc2(nn34)
      Real*8 A(nn1,nn2),B(nn1,nn34),C(nn34,nn2),D(nn34,nn34),tot,
     +       Z(nn34,nn34),DD(nn34,nn34),Dmin(nn34,nn34),Dmoy(nn34,nn34),
     +       Dmax(nn34,nn34),Prob(nn34),Dligne(nn34),Dcol(nn34),
     +       count(nn34,nn34),dseed
      Common A,B,C,D,tot,Z,DD,Dmin,Dmoy,Dmax,Prob,Dligne,Dcol,count,
     +       dseed,EspD,n1,n2,n3,n4,DLT,DEQ,permod,
     +       methode,nperm,nclasse,langue
      Common /stand/ bbar,sb,cbar,sc
      Common /pourperm/ Aprime,iordre
C      Character nom*10
C Calculations for method 1 (contingency table) or 2 (correlation)
C
      if(ipar.eq.0) then
         do 2 i=1,n1
         do 2 j=1,n2
    2    Aprime(j,i)=A(i,j)
         else
         call Permute
         endif
C      nom=' Aprime = '
C      if(ipar.gt.0) call Imprime(Aprime,nn2,nn1,n2,n1,nom)
C Standardize B and/or C ?
C First, the case of contingency tables
      if(methode.eq.1) then
         do 4 k=1,n4
         bbar(k)=0.0
    4    sb(k)=1.0
         do 6 k=1,n3
         cbar(k)=0.0
    6    sc(k)=1.0
      else
C Next, the case of correlations (methode.eq.2).
C If permutations are done using methods 1 or 2, and if computation method
C is 2 (correlation), standardize B the first time only. Reason: in that
C case, the permutations always leave the same number of occurrences for 
C each value of B.
      if((permod.ge.3).or.(ipar.eq.0)) then
         do 8 k=1,n4
         sb(k)=0.0
    8    sb2(k)=0.0
         do 12 i=1,n1
         do 12 j=1,n2
         if(Aprime(j,i).eq.1.0) then
            do 10 k=1,n4
            sb(k)=sb(k)+B(i,k)
            sb2(k)=sb2(k)+B(i,k)*B(i,k)
   10       continue
            endif
   12    continue
         do 14 k=1,n4
         bbar(k)=sb(k)/tot
         sb(k)=dsqrt((sb2(k)-((sb(k))**2/tot))/(tot-1.0))
   14    continue
C Divide by TOT because that is the number of rows of the inflated matrix.
C sb = standard deviation of each variable b in inflated matrix; bbar = mean.
C         write(*,*) 'sb',(sb(i), i=1,n4)
         endif
C For method 2 (correlation), standardize C
      do 16 k=1,n3
      sc(k)=0.0
   16 sc2(k)=0.0
      do 20 i=1,n1
      do 20 j=1,n2
      if(Aprime(j,i).eq.1.0) then
         do 18 k=1,n3
         sc(k)=sc(k)+C(k,j)
         sc2(k)=sc2(k)+C(k,j)*C(k,j)
   18    continue
         endif
   20 continue
      do 22 k=1,n3
      cbar(k)=sc(k)/tot
      sc(k)=dsqrt((sc2(k)-((sc(k))**2/tot))/(tot-1.0))
   22 continue
C Divide by TOT because that is the number of rows of the inflated matrix.
C sc = standard deviation of each variable c in inflated matrix; cbar = mean.
C      write(*,*) 'sc',(sc(i), i=1,n3)
      endif
C Compute D = C A' B
      call Product1(C,Aprime,Caprime,n3,n2,n1)
C      nom=' Caprime  '
C      call Imprime(Caprime,nn34,nn1,n3,n1,nom)
      if(ipar.eq.0) then
         call Product2(Caprime,B,D,n3,n1,n4,tot,methode)
C      nom=' D        '
C      call Imprime(D,nn34,nn34,n3,n4,nom)
         else
         call Product2(Caprime,B,DD,n3,n1,n4,tot,methode)
C      nom=' DD       '
C      call Imprime(DD,nn34,nn34,n3,n4,nom)
         endif
      return
      end

      Subroutine Product1(mat1,mat2,mat3,n3,n2,n1)
      Parameter (nn1=300,nn2=30,nn34=30)
      Real*8 mat1(nn34,nn2),mat2(nn2,nn1),mat3(nn34,nn1),val,
     +       bbar(nn34),sb(nn34),cbar(nn34),sc(nn34)
      Common /stand/ bbar,sb,cbar,sc
C Product C A'
      do 30 i=1,n1
      do 20 k=1,n3
      if(sc(k).eq.0.0) then
         mat3(k,i)=0.0
         goto 20
         endif
      val=0.
      do 10 j=1,n2
   10 val=val+mat2(j,i)*(mat1(k,j)-cbar(k))/sc(k)
      mat3(k,i)=val
   20 continue
   30 continue
      return
      end

      Subroutine Product2(mat1,mat2,mat3,n3,n1,n4,tot,methode)
      Parameter (nn1=300,nn2=30,nn34=30)
      Real*8 mat1(nn34,nn1),mat2(nn1,nn34),mat3(nn34,nn34),val,
     +       bbar(nn34),sb(nn34),cbar(nn34),sc(nn34),tot
      Common /stand/ bbar,sb,cbar,sc
C Product (C A') B
C
      do 30 k=1,n3
      do 20 i=1,n4
      if(sb(i).eq.0.0) then
         mat3(k,i)=0.0
         goto 20
         endif
      val=0.
      do 10 j=1,n1
   10 val=val+mat1(k,j)*(mat2(j,i)-bbar(i))/sb(i)
      if(methode.eq.2) val=val/(tot-1.0)
	  mat3(k,i)=val
   20 continue
   30 continue
C      write(*,100) tot, bbar(1),sb(1),cbar(1),sc(1)
      return
  100 format(f5.0,4f10.5)
      end

      Subroutine Prod2(ipar,F)
      Parameter (nn1=300,nn2=30,nn34=30)
      Integer EspD(nn34,nn34),n1,n2,n3,n4,DLT(nn34,nn34),DEQ(nn34,nn34),
     +        permod
      Integer iordre(nn1)
      Real*8 Aprime(nn2,nn1),sx,sx2,sy(nn34),sy2(nn34),temp,
     +       F,ScTotal,ScIntra
      Real*8 A(nn1,nn2),B(nn1,nn34),C(nn34,nn2),D(nn34,nn34),tot,
     +       Z(nn34,nn34),DD(nn34,nn34),Dmin(nn34,nn34),Dmoy(nn34,nn34),
     +       Dmax(nn34,nn34),Prob(nn34),Dligne(nn34),Dcol(nn34),
     +       count(nn34,nn34),dseed
      Common A,B,C,D,tot,Z,DD,Dmin,Dmoy,Dmax,Prob,Dligne,Dcol,count,
     +       dseed,EspD,n1,n2,n3,n4,DLT,DEQ,permod,
     +       methode,nperm,nclasse,langue
      Common /pourperm/ Aprime,iordre
C Calculations for methods 3 to 6
C
      if(ipar.eq.0) then
         do 2 i=1,n1
         do 2 j=1,n2
    2    Aprime(j,i)=A(i,j)
         else
         call Permute
         endif
C      nom=' Aprime = '
C      if(ipar.gt.0) call Imprime(Aprime,nn2,nn1,n2,n1,nom)
      sx=0.0
      sx2=0.0
      goto (99,99,3,4,5,6) methode
C Method 3: D(1xn4)
    3 if(n3.gt.1) stop 'Meth. 3: n3 > 1'
      do 20 i=1,n4
      count(1,i)=0
      sy(i)=0.0
   20 sy2(i)=0.0
      do 22 i=1,n1
      do 22 j=1,n2
      if(Aprime(j,i).eq.1.0) then
         sx=sx+C(1,j)
         sx2=sx2+C(1,j)*C(1,j)
         do 21 mm=1,n4
         if(B(i,mm).eq.0) goto 21
         sy(mm)=sy(mm)+C(1,j)
         sy2(mm)=sy2(mm)+C(1,j)*C(1,j)
         count(1,mm)=count(1,mm)+1
   21    continue
         endif
   22 continue
      ScTotal=sx2-(sx**2/tot)
      ScIntra=0.0
      do 24 k=1,n4
      if(count(1,k).le.1) goto 24
      temp=sy2(k)-(sy(k)**2/count(1,k))
      if(ipar.eq.0) then
C I finally decided to use the ratio of sum of squares instead of the 
C ratio of variances; the latter is not upper-bounded by 1. See Ecology paper.
C         D(1,k)=temp*(tot-1.0)/ScTotal/(count(k,1)-1.0)
         D(1,k)=temp/ScTotal
         else
C         DD(1,k)=temp*(tot-1.0)/ScTotal/(count(k,1)-1.0)
         DD(1,k)=temp/ScTotal
         endif      
      ScIntra=ScIntra+temp
   24 continue
      kk=0
      do 26 i=1,n4
   26 if(count(1,i).gt.1) kk=kk+1
      goto 98
C Method 4: D(n3x1)
    4 if(n4.gt.1) stop 'Meth. 4: n4 > 1'
      do 30 i=1,n3
      count(i,1)=0
      sy(i)=0.0
   30 sy2(i)=0.0
      do 32 i=1,n1
      do 32 j=1,n2
      if(Aprime(j,i).eq.1.0) then
         sx=sx+B(i,1)
         sx2=sx2+B(i,1)*B(i,1)
         do 31 mm=1,n3
         if(C(mm,j).eq.0) goto 31
         sy(mm)=sy(mm)+B(i,1)
         sy2(mm)=sy2(mm)+B(i,1)*B(i,1)
         count(mm,1)=count(mm,1)+1
   31    continue
         endif
   32 continue
      ScTotal=sx2-(sx**2/tot)
      ScIntra=0.0
      do 34 k=1,n3
      if(count(k,1).le.1) goto 34
      temp=sy2(k)-(sy(k)**2/count(k,1))
      if(ipar.eq.0) then
C         D(k,1)=temp*(tot-1.0)/ScTotal/(count(k,1)-1.0)
         D(k,1)=temp/ScTotal
         else
C         DD(k,1)=temp*(tot-1.0)/ScTotal/(count(k,1)-1.0)
         DD(k,1)=temp/ScTotal
         endif      
      ScIntra=ScIntra+temp
   34 continue
      kk=0
      do 36 i=1,n3
   36 if(count(i,1).gt.1) kk=kk+1
      goto 98
C Method 5: D(1xn4)
    5 if(n3.gt.1) stop 'Meth. 5: n3 > 1'
      if(n4.gt.1) stop 'Meth. 5: n4 > 1'
      do 40 i=1,nclasse
      count(1,i)=0
      sy(i)=0.0
   40 sy2(i)=0.0
      do 42 i=1,n1
      do 42 j=1,n2
      if(Aprime(j,i).eq.1.0) then
         sx=sx+C(1,j)
         sx2=sx2+C(1,j)*C(1,j)
         mm=B(i,1)
         sy(mm)=sy(mm)+C(1,j)
         sy2(mm)=sy2(mm)+C(1,j)*C(1,j)
         count(1,mm)=count(1,mm)+1
         endif
   42 continue
      ScTotal=sx2-(sx**2/tot)
      ScIntra=0.0
      do 44 k=1,nclasse
      if(count(1,k).le.1) goto 44
      temp=sy2(k)-(sy(k)**2/count(1,k))
      if(ipar.eq.0) then
C         D(1,k)=temp*(tot-1.0)/ScTotal/(count(k,1)-1.0)
         D(1,k)=temp/ScTotal
         else
C         DD(1,k)=temp*(tot-1.0)/ScTotal/(count(k,1)-1.0)
         DD(1,k)=temp/ScTotal
         endif      
      ScIntra=ScIntra+temp
   44 continue
      kk=0
      do 46 i=1,nclasse
   46 if(count(1,i).gt.1) kk=kk+1
      goto 98
C Method 6: D(n3x1)
    6 if(n3.gt.1) stop 'Meth. 6: n3 > 1'
      if(n4.gt.1) stop 'Meth. 6: n4 > 1' 
      do 50 i=1,nclasse
      count(i,1)=0
      sy(i)=0.0
   50 sy2(i)=0.0
      do 52 i=1,n1
      do 52 j=1,n2
      if(Aprime(j,i).eq.1.0) then
         sx=sx+B(i,1)
         sx2=sx2+B(i,1)*B(i,1)
         mm=C(1,j)
         sy(mm)=sy(mm)+B(i,1)
         sy2(mm)=sy2(mm)+B(i,1)*B(i,1)
         count(mm,1)=count(mm,1)+1
         endif
   52 continue
      ScTotal=sx2-(sx**2/tot)
      ScIntra=0.0
      do 54 k=1,nclasse
      if(count(k,1).le.1) goto 54
      temp=sy2(k)-(sy(k)**2/count(k,1))
      if(ipar.eq.0) then
C         D(k,1)=temp*(tot-1.0)/ScTotal/(count(k,1)-1.0)
         D(k,1)=temp/ScTotal
         else
C         DD(k,1)=temp*(tot-1.0)/ScTotal/(count(k,1)-1.0)
         DD(k,1)=temp/ScTotal
         endif      
      ScIntra=ScIntra+temp
   54 continue
      kk=0
      do 56 i=1,nclasse
   56 if(count(i,1).gt.1) kk=kk+1
C
   98 if(kk.eq.1) then
         F=0.0
         else
         F=(ScTotal-ScIntra)*(tot-dfloat(kk))/ScIntra/dfloat(kk-1)
         endif
C      write(8,110) ScTotal,ScIntra,(ScTotal-ScIntra),kk,F
   99 return
  110 format(' ScTotal =',f10.5,'  ScIntra =',f10.5,
     +       '  ScInter =',f10.5,'  kk =',i3,' F =',f10.5)
      end

      Subroutine Test(G,Chi2P,ipar)
      Parameter (nn1=300,nn2=30,nn34=30)
      Integer EspD(nn34,nn34),n1,n2,n3,n4,DLT(nn34,nn34),DEQ(nn34,nn34),
     +        permod
      Real*8 Dtot,Dobs,G,Chi2P,Desp,Diff
      Real*8 A(nn1,nn2),B(nn1,nn34),C(nn34,nn2),D(nn34,nn34),tot,
     +       Z(nn34,nn34),DD(nn34,nn34),Dmin(nn34,nn34),Dmoy(nn34,nn34),
     +       Dmax(nn34,nn34),Prob(nn34),Dligne(nn34),Dcol(nn34),
     +       count(nn34,nn34),dseed
      Common A,B,C,D,tot,Z,DD,Dmin,Dmoy,Dmax,Prob,Dligne,Dcol,count,
     +       dseed,EspD,n1,n2,n3,n4,DLT,DEQ,permod,
     +       methode,nperm,nclasse,langue
C "G test" on contingency table
C
      nvides=0
      do 30 i=1,n3
   30 Dligne(i)=0.0
      do 32 j=1,n4
   32 Dcol(j)=0.0
      Dtot=0.0
      G=0.0
      Chi2P=0.0
      if(ipar.eq.0) then
         do 42 i=1,n3
         do 42 j=1,n4
         Dtot=Dtot+D(i,j)
         Dligne(i)=Dligne(i)+D(i,j)
   42    Dcol(j)  =Dcol(j)  +D(i,j)
         else
         do 44 i=1,n3
         do 44 j=1,n4
         Dtot=Dtot+DD(i,j)
         Dligne(i)=Dligne(i)+DD(i,j)
   44    Dcol(j)  =Dcol(j)  +DD(i,j)
         endif
      if((ipar.eq.0).and.(dabs(Dtot-tot).gt.0.000001)) then
         if(langue.eq.1) then
            write(8,120) tot,Dtot
            else
            write(8,220) tot,Dtot
            endif
         endif
      do 48 i=1,n3
      do 48 j=1,n4
      if(ipar.eq.0) then
         Dobs=D(i,j)
         else
         Dobs=DD(i,j)
         endif
      Desp=Dligne(i)*Dcol(j)/Dtot
      if(ipar.gt.0) goto 46
      if(Dobs.ge.5.0) then
         Diff=Dobs-Desp
         if(Diff.lt.0.0) Diff=-Diff
         Z(i,j)=Diff/dsqrt(Dobs*(1.0-(Dobs/Dtot)))
         else
         Z(i,j)=-9.0
         endif
   46 if(Desp.eq.0.0) goto 50
      Chi2P=Chi2P+(((Dobs-Desp)**2)/Desp)
      if(Dobs.ne.0.0) then
         G=G+2.0*(Dobs*dlog(Dobs/Desp))
         else
         nvides=nvides+1
         endif
   48 continue
      goto 52
   50 Chi2P=0.0
      G=0.0
   52 continue
      if((ipar.eq.0).and.(nvides.gt.0)) then
         if(langue.eq.1) then
            write(8,121) nvides
            else
            write(8,221) nvides
            endif
         endif
      return
  120 format(' *** Attention: Tot = ',f6.0,' alors que Dtot = ',f6.0)
  220 format(' *** Warning: Tot = ',f6.0,' while Dtot = ',f6.0)
  121 format(' Nombre de cases vides (Obs=0) dans D =',i4)
  221 format(' Number of empty cells (Obs=0) in D =',i4)
      end
 
      Subroutine Imprime(A,nn1,nn2,n1,n2,nom)
C Prints matrices. For debugging purpose.
C
      Real*8 A(nn1,nn2)
      Character nom*10
      write(8,100) nom
      do 10 i=1,n1
      write(8,101) (A(i,j), j=1,n2)
   10 continue
      write(8,*)
      return
  100 format(1x,a10)
  101 format(8f10.5)
      end

      Subroutine Ggper2 (dseed,k)
C This subroutine permutes the first 'k' values of vector 'iordre' at random
C in an equiprobable way. This property has been checked through intensive 
C simulations.
C
      Parameter (nn1=300,nn2=30,nn34=30)
      Integer k,i,itemp,km1,m,iordre(nn1)
      Real*8 Aprime(nn2,nn1),dseed
      Common /pourperm/ Aprime,iordre
      m=k
      km1=k-1
      do 10 i=1,km1
         j = 1+ggubfs(dseed)*m
         itemp = iordre(m)
         iordre(m) = iordre(j)
         iordre(j) = itemp
         m=m-1
   10 continue
      return
      end

      REAL FUNCTION GGUBFS (DSEED)
      DOUBLE PRECISION   DSEED
      DOUBLE PRECISION   D2P31M,D2P31
C                                  D2P31M=(2**31) - 1
      DATA               D2P31M/2147483647.D0/
      DATA               D2P31 /2147483648.D0/
      DSEED = DMOD(16807.D0*DSEED,D2P31M)
      GGUBFS = DSEED / D2P31
      RETURN
      END
